open Ast

module NameMap = Map.Make(struct
  type t = string
  let compare x y = Pervasives.compare x y
end)

exception ReturnException of string * string NameMap.t

(* Main entry point: run a program *)

let run (vars, funcs) =
  (* Put function declarations in a symbol table *)
  let func_decls = List.fold_left
      (fun funcs fdecl -> NameMap.add fdecl.fname fdecl funcs)
      NameMap.empty funcs
  in
  
	(* Function to trim quotes from strings *)
	let rec trimQuote string =
	let l = String.length string in
	if string.[0]='"' then
	trimQuote (String.sub string 1 (l-1))
	else if string.[l-1]='"' then
	trimQuote (String.sub string 0 (l-1))
	else string
	
	in
	 
  (* Invoke a function and return an updated global symbol table *)
  let rec invoke fdecl actuals globals =
    (* Evaluate an expression and return (value, updated environment) *)
    let rec eval env = function
				Int_Literal(i) -> i, env
			| String_Literal(i) -> i, env
      | Noexpr -> "1", env (* must be non-zero for the for loop predicate *)
      | Id(var) ->
	  		let locals, globals = env in
	  		if NameMap.mem var locals then
	    	(NameMap.find var locals), env
	  		else if NameMap.mem var globals then
	    	(NameMap.find var globals), env
	  		else raise (Failure ("undeclared identifier in Id " ^ var))
      | Assign(var, e) ->
	  			let v, (locals, globals) = eval env e in
	  			if NameMap.mem var locals then
	    		v, (NameMap.add var v locals, globals)
	  			else if NameMap.mem var globals then
	    		v, (locals, NameMap.add var v globals)
	  			else raise (Failure ("undeclared identifier in Assign " ^ var))
			| IntAssign(var, e) ->
	  			let v, (locals, globals) = eval env e in
	  			if NameMap.mem var locals then
	    		v, (NameMap.add var v locals, globals)
	  			else if NameMap.mem var globals then
	    		v, (locals, NameMap.add var v globals)
	  			else raise (Failure ("undeclared identifier in Assign " ^ var))
			| LinkAssign(var, e, e2) ->
	  			let v, (locals, globals) = eval env e in
					let v2, (locals, globals) = eval env e2 in
					let trimV2 = trimQuote v2 in
	  			if NameMap.mem var locals then
	    		v, (NameMap.add var ("<a href=" ^ v ^ ">" ^ trimV2 ^ "</a><br>") locals, globals)
	  			else if NameMap.mem var globals then
	    		v, (locals, NameMap.add var ("<a href=" ^ v ^ ">" ^ trimV2 ^ "</a><br>") globals)
	  			else raise (Failure ("undeclared identifier in LinkAssign " ^ var))	
			| FormAssign(var, methodName, action) -> 
				  let v, (locals, globals) = eval env methodName in	
					let v2,(locals, globals) = eval env action in				
	  			if NameMap.mem var locals then
	    		v, (NameMap.add var ("<form onsubmit=\"return true\" method=" ^ v
					 ^ " action=" ^ v2 ^ ">") locals, globals)
	  			else if NameMap.mem var globals then
	    		v, (locals, NameMap.add var ("<a href=" ^ v ^ "dotcom") globals)
	  			else raise (Failure ("undeclared identifier in LinkAssign " ^ var))
			| InputAssign(var, label, size, name) -> 
				  let labelStr, (locals, globals) = eval env label in
						let trimmedLabelStr = trimQuote labelStr in	
					let sizeStr,(locals, globals) = eval env size in				
					let nameStr,(locals, globals) = eval env name in
	  			if NameMap.mem var locals then
	    		trimmedLabelStr, (NameMap.add var ("<p>" ^ trimmedLabelStr
					 ^ " <input type=\"text\" maxlength=" ^ sizeStr ^ " name=" 
					 ^ nameStr ^ " size=" ^ sizeStr ^ ">") locals, globals)
	  			else if NameMap.mem var globals then
	    		trimmedLabelStr, (locals, NameMap.add var ("<a href=" ^ trimmedLabelStr ^ "dotcom") globals)
	  			else raise (Failure ("undeclared identifier in LinkAssign " ^ var))
			| ImageAssign(var, src, width, height) ->
				let srcStr, (locals, globals) = eval env src in					
					let widthStr,(locals, globals) = eval env width in				
					let heightStr,(locals, globals) = eval env height in
	  			if NameMap.mem var locals then
	    		srcStr, (NameMap.add var ("<img src=" ^ srcStr
					 ^ " width=" ^ widthStr ^ " height=" 
					 ^ heightStr ^ ">") locals, globals)
	  			else if NameMap.mem var globals then
	    		srcStr, (locals, NameMap.add var ("<img src=" ^ srcStr
					 ^ " width=" ^ widthStr ^ " height=" 
					 ^ heightStr ^ ">") globals)
	  			else raise (Failure ("undeclared identifier in LinkAssign " ^ var))
			| Invoke ("bold", [e]) ->
				 let v, env = eval env e in 		
				 let vStr = trimQuote v in 		
				 print_endline("<b>" ^ vStr ^ "</b><br>");
				 "0", env
			| Button (bType, value) -> 
				let bTypeStr, (locals, globals) = eval env bType in
				let valueStr, (locals, globals) = eval env value in 
				"<input type=" ^ bTypeStr ^ " value=" ^ valueStr ^ ">", (NameMap.add "dummyKey" ("") locals, globals)
			| Invoke ("close", [e]) ->
				let v, env = eval env e in 
				print_endline("</body>\n</html>");
				"0", env 
			| Invoke ("open", [e]) -> 
				let v, env = eval env e in 
				print_endline("<html>\n<body>");
				"0", env 	
			| Invoke ("header1", [e]) -> 
				let v, env = eval env e in
				let vTrim = trimQuote v in  
				print_endline("<h1>" ^ vTrim ^ "</h1>");
				"0", env  
			| Invoke ("header2", [e]) -> 
				let v, env = eval env e in
				let vTrim = trimQuote v in  
				print_endline("<h2>" ^ vTrim ^ "</h2>");
				"0", env
			| Invoke ("header3", [e]) -> 
				let v, env = eval env e in
				let vTrim = trimQuote v in  
				print_endline("<h3>" ^ vTrim ^ "</h3>");
				"0", env    
			| Invoke ("header4", [e]) -> 
				let v, env = eval env e in
				let vTrim = trimQuote v in  
				print_endline("<h4>" ^ vTrim ^ "</h4>");
				"0", env  	
			| Invoke ("header5", [e]) -> 
				let v, env = eval env e in
				let vTrim = trimQuote v in  
				print_endline("<h5>" ^ vTrim ^ "</h5>");
				"0", env  	
			| Invoke ("header6", [e]) -> 
				let v, env = eval env e in
				let vTrim = trimQuote v in  
				print_endline("<h6>" ^ vTrim ^ "</h6>");
				"0", env  	
			| Invoke ("para", [e]) -> 
				let v, env = eval env e in
				let vTrim = trimQuote v in  
				print_endline("<p>" ^ vTrim ^ "</p>");
				"0", env  	
			| Invoke ("text", [e]) -> 
				let v, env = eval env e in
				print_endline("<br>" ^ v ^ "</br>");
				"0", env
			| Invoke ("comment", [e]) -> 
				let v, env = eval env e in
				let vTrim = trimQuote v in 
				print_endline("<!--" ^ vTrim ^ "-->");
				"0", env
			| Invoke ("title", [e]) -> 
				let v, env = eval env e in
				let vTrim = trimQuote v in 
				print_endline("<title>" ^ vTrim ^ "</title>");
				"0", env
	    | Invoke ("insert", [e]) ->
	  		let v, env = eval env e in
				if (v = "\"line\"") then 
					print_endline("<hr>")
				else
	  		print_endline (v);
	  		"0", env
      | Invoke (f, actuals) ->
	  		let fdecl =
	    	try NameMap.find f func_decls
	    	with Not_found -> raise (Failure ("undefined function " ^ f))
	  		in
	  		let actuals, env = List.fold_left
	      (fun (actuals, env) actual ->
				let v, env = eval env actual in v :: actuals, env)
   	      ([], env) actuals
	  			in
	  			let (locals, globals) = env in
	  			try
	    			let globals = invoke fdecl actuals globals in "0", (locals, globals)
	  				with ReturnException(v, globals) -> v, (locals, globals)
    			in

    (* Execute a statement and return an updated environment *)
    let rec exec env = function
				Block(stmts) -> List.fold_left exec env stmts
      | Expr(e) -> let _, env = eval env e in env
      | Return(e) ->
			  let v, (locals, globals) = eval env e in
			  raise (ReturnException(v, globals))
		    in
		    (* Enter the function: bind actual values to formal arguments *)
		    let locals =
      	try List.fold_left2
	  		(fun locals formal actual -> NameMap.add formal actual locals)
	  		NameMap.empty fdecl.formals actuals
      	with Invalid_argument(_) ->
				raise (Failure ("wrong number of arguments passed to " ^ fdecl.fname))
    		in
    		(* Initialize local variables to 0 *)
  		  let locals = List.fold_left
				(fun locals local -> NameMap.add local "0" locals) locals fdecl.locals
    		in
    		(* Execute each statement in sequence, return updated global symbol table *)
    		snd (List.fold_left exec (locals, globals) fdecl.body)

  			(* Run a program: initialize global variables to 0, find and run "page" *)
  			in let globals = List.fold_left
      	(fun globals vdecl1 -> NameMap.add vdecl1 "0" globals) NameMap.empty vars
  			in try
    		invoke (NameMap.find "run" func_decls) [] globals
  			with Not_found -> raise (Failure ("did not find the run() function"))
